{-# OPTIONS_GHC -F -pgmF htfpp #-}
module FOLProvTest where
import FOLProv
import FOLData
import Parser
import qualified Data.HashMap.Strict as Map

import Test.Framework


test_subsumes :: IO()
test_subsumes = do
  assertEqual False (subsumes (fsOf a) (fsOf b))
  assertEqual False (subsumes (fsOf b) (fsOf a)) 
  assertEqual False (subsumes (fsOf a) (fsOf c))
  assertEqual True (subsumes (fsOf a) (fsOf d))
  assertEqual True (subsumes (fsOf a) (fsOf e))
  assertEqual True (subsumes (fsOf c) (fsOf d))
  assertEqual False (subsumes (fsOf d) (fsOf c))
  assertEqual True (subsumes (fsOf f) (fsOf g))
  assertEqual False (subsumes (fsOf g) (fsOf f))

test_unification :: IO()
test_unification = do
  assertEqual (Just uniAB) (unificator (head (fsOf a)) (head (fsOf b)))
  assertEqual (Just (reverse uniAB)) (unificator (head (fsOf b)) (head (fsOf a))) 

test_condensation :: IO()
test_condensation = do
  assertEqual False (condensateOf (fsOf a) (fsOf c))
  assertEqual False (condensateOf (fsOf a) (fsOf d)) -- a proper multifactor is needed so the term itself is not allowed
  assertEqual True (condensateOf (fsOf c) (fsOf d))
  assertEqual False (condensateOf (fsOf d) (fsOf c)) 
  
test_resolution :: IO()
test_resolution = do
  assertEqual False (testString s0)
  assertEqual False (testString s0')
  assertEqual False (testString s1)

test_resolution2 :: IO()
test_resolution2 = do
    assertEqual False (testString s2)
    assertEqual True (testString s3)


a,b,c,d,e,f,g :: Clause

a = clauseOf "P(x,y)"
b = clauseOf "P(y,F(A))"
c = clauseOf "P(y,x) or P(y,y)"
d = clauseOf "P(A,A) or P(A,B)"
e = clauseOf "P(x,y) or P(y,y)"
f = clauseOf "Q(x) or Q(F(y))"
g = clauseOf "Q(x) or Q(F(x))"
h = clauseOf "(~ P(C,v1)) or (~ P(v1,C)) or (~ P(v1,F(v1))) or P(F(v1),v1)"
i = clauseOf "(~ P(C,v1)) or (~ P(v1,C)) or P(F(v1),v1)"
j = clauseOf "(~ P(v2,v1)) or (~ P(v1,v2)) or P(F(v1),v1)"
k = clauseOf "Q(F(y))"
l = clauseOf "~Q(F(x)) or ~Q(x)"

-- h [~(P(c,1)),~(P(1,c)),~(P(1,f(1))),P(f(1),1)] 
-- i > [~(P(c,1)),~(P(1,c)),P(f(1),1)] 
-- j > [~(P(2,1)),~(P(1,2)),P(f(1),1)]


s0,s0',s1,s2,s3 :: String
s0 = "P(x) and ~(P(y))"
s0' = "~(P(x)) and P(y)"
s1 = "(?y. !z.((~ P(z,y) or ~( ?x.(P(z,x) and P(x,z)))) and (?x.(P(z,x) and P(x,z)) or P(z,y))))"
s2 = "(P(x) or Q(x)) and (not P(x) or Q(F(y))) and (P(x) or not Q(F(x)))\
     \and (not P(x) or not Q(x))"
s3 = "(P(x) or Q(x)) and (not P(x) or Q(F(x))) and (P(x) or not Q(F(x)))\
     \and (not P(x) or not Q(x))"


uniAB:: [(Term,Term)]
uniAB = [(Var "x",Func "F" [Const "A"]),(Var "y",Func "F" [Const "A"])]

fsOf :: Clause -> [Formula]
fsOf = clauseFormulas
    
clauseOf:: String -> Clause
clauseOf = fst . head . filter cond . Map.elems . cnf . parseFOL
  where
    cond (Clause fs _ _ _,_) = any cond' fs
    cond' (Pred p _) = p /= equalPredicate
    cond' (Not (Pred p _)) = p /= equalPredicate
  