{-# LANGUAGE DeriveGeneric #-}

module FOLData where

import           Data.Hashable
--import qualified Data.HashSet  as Set
import           GHC.Generics  (Generic)

type Constant = String

type Function = String

type Predicate = String

equalPredicate :: Predicate
equalPredicate = "_EQUAL"

type Variable = String


data Connective
     = And
     | Or
     | Imp
     deriving (Eq,Generic)


data Quantifier
     = ForAll
     | Exists
     deriving (Eq,Generic)

data Term
     = Const Constant
     | Var Variable
     | Func Function [Term]
     deriving (Eq,Generic)

data Formula
     = Pred Predicate [Term]
     | Not Formula
     | Conn Formula Connective Formula
     | Quant Quantifier Variable Formula
     | Box
     deriving (Eq,Generic)

data Clause = Clause { clauseFormulas :: [Formula]
                     , clauseHash :: Int
                     , parent1 :: Maybe Clause
                     , parent2 :: Maybe Clause}
            deriving (Generic)

instance Eq Clause where
  (==) (Clause _ h1 _ _) (Clause _ h2 _ _) = h1 == h2

instance Hashable Clause where
   hashWithSalt s (Clause fs _ _ _) = s `hashWithSalt` fs 

instance Hashable Connective

instance Hashable Quantifier

instance Hashable Term

instance Hashable Formula


instance Show Connective where
  show And = "/\\"
  show Or = "\\/"
  show Imp = "==>"

instance Show Quantifier where
  show ForAll = "!"
  show Exists = "?"

instance Show Term where
  show (Const x) = "c" ++ x
  show (Var x) = "v" ++ x
  show (Func x ts) = "f"++ x ++ "(" ++ flattenListString (show ts) ++ ")"

instance Show Formula where
  show (Pred p ts) = ""++ p ++ "(" ++ flattenListString (show ts) ++ ")"
  show (Not f) = "~("++ show f ++ ")"
  show (Conn f1 c f2) = "(" ++ show f1 ++ show c ++ show f2 ++ ")"
  show (Quant q v f) = show q ++ "v"++flattenListString (show v) ++ "." ++ show f
  show (Box) = "Box"

flattenListString :: String -> String
flattenListString = filter (`notElem` "[]\"")

instance Show Clause where
  show (Clause fs _ p1 p2) = (show fs ++ "-->["++"\nParent1:"++show p1++",\nParent2:"++show p2++"\n]")

