
{-# LANGUAGE BangPatterns #-}
module FOLProv where

import           Data.Hashable (hash)
import qualified Data.HashMap.Strict as Map
import qualified Data.HashSet        as Set
import qualified Data.List           as List
import           Parser
--import           TPTPtoFOL
import           FOLData

import           Debug.Trace

type CNFMap = Map.HashMap Int (Clause, Set.HashSet Int)



cnfMapToString:: CNFMap -> String
cnfMapToString cnfmap = unlines (map (showcl . fst) el)
  where
    showcl cl = show (clauseFormulas cl) ++ " |-> " ++ parents cl
    parents cl = parent (parent1 cl) ++" | "++parent (parent2 cl)
    parent (Just cl) = show $ clauseFormulas cl
    parent Nothing = "N"
    el = Map.elems cnfmap

{- test formula -}

testString:: String -> Bool
testString s = testFormula . cnf . parseFOL $ s

testFormula:: CNFMap -> Bool
testFormula = testFormula' -- . addEqualityAxioms

testFormula':: CNFMap -> Bool
testFormula' cnfmap = let
  cnfmap' = resolutionSteps cnfmap
  in
     if boxCheck cnfmap'
     then trace ("Box:" ++ show (Map.size cnfmap')) $ -- ++ "\n" ++ cnfMapToString cnfmap') $ 
       --trace (show (Map.lookup (clauseHash boxClause) cnfmap'))$ 
       False -- "Box found. " ++ show (Map.size cnfmap') ++ " Clauses in map." 
     else if cnfmap == cnfmap'
          then True -- "All possibilities checked, no Box found."
          else testFormula' cnfmap'

{- condensation -}

mapCondensateOf :: [Formula] -> CNFMap -> Bool
mapCondensateOf fs = Map.foldr (\(cl,_) res -> res || clauseFormulas cl `condensateOf` fs) False


condensateOf :: [Formula] -> [Formula] -> Bool
condensateOf f1 f2 =
  let
    f1Facts = factorsOf f1
  in --trace (show f1 ++"|"++show f2++"|"++ show f1Facts) $
     any (`subsumes` f2) f1Facts --  || sub' f1 f2
  where
    sub' f1 f2 = 
      if length f1 < length f2
      then subsumes f1 f2
      else False
{-- 
  case subsumes' f1 f2 Map.empty f2 (varsOf f2) of
    Just _ ->
      if length f1 < length f2
      then True
      else False
    Nothing -> False
--}
    
factorsOf :: [Formula] -> [[Formula]]
factorsOf fs = factorsOf' fs []

factorsOf' :: [Formula] -> [[Formula]] -> [[Formula]]
factorsOf' fs fcs = foldr (\f facts -> factorsOfFormula f fs facts) fcs fs
  where
    factorsOfFormula :: Formula -> [Formula] -> [[Formula]] -> [[Formula]]
    factorsOfFormula form forms facts = let
        forms' = filter (/= form) forms
      in if form `isFactorOf` forms'
         then if forms `elem` facts
              then factorsOf' forms' facts
              else factorsOf' forms' (forms:facts)
         else facts
    isFactorOf :: Formula -> [Formula] -> Bool
    isFactorOf f fs = [f] `subsumes` fs



{- subsumption -}

backwardSubsumption :: [Formula] -> CNFMap -> CNFMap
backwardSubsumption fs cnfmap = foldr (\i !m' -> Map.delete i m') cnfmap subsumedInd
  where
    subsumedInd = Map.foldr (\(e,_) !acc-> func fs e acc) [] cnfmap
    func fs e acc =
      if fs `subsumes` (clauseFormulas e) 
      then clauseHash e : acc
      else acc
      

mapSubsumes :: [Formula] -> CNFMap -> Bool
mapSubsumes fs = Map.foldr (\(cl,_) res -> res || clauseFormulas cl `subsumes` fs) False

subsumes :: [Formula] -> [Formula] -> Bool
subsumes f1 f2 =
  case subsumes' f1 f2 Map.empty f2 (varsOf f2) of
    Just _ -> --trace ("true: "++show f1 ++"|"++show f2) $
      True
    Nothing -> --trace ("false: "++show f1 ++"|"++show f2) $
     False
  where

subsumes' :: [Formula] -> [Formula] -> Map.HashMap Term Term 
             -> [Formula] -> Set.HashSet Variable -> Maybe (Map.HashMap Term Term)
subsumes' [] _ subs _ _ = Just subs -- all predicates are fixed
subsumes' (_:_) [] _ _ _ = Nothing -- nor mor option to fix a predicate
subsumes' (p1:ps1) (p2:ps2) subs fs2 vars = 
  --trace ("subs'"++show (p1:ps1) ++" | " ++ show (p2:ps2) ++ " | " ++ show subs) $
  case buildSubsF (p1,p2) subs vars of
    Just subs' -> --trace ("next call: "++show subs' ++ show ps1 ++ show fs2') $
        case subsumes' ps1 fs2' subs' fs2' vars of
          Just s -> Just s
                 -- a subtituion was found
          Nothing -> subsumes' (p1:ps1) ps2 subs fs2 vars
                  -- try to fix Predicate in an other way
    Nothing -> -- trace ("No buildSubsF") $
      -- could not match predicate with subtitution
      subsumes' (p1:ps1) ps2 subs fs2 vars
  where
    fs2' = List.delete p2 fs2


buildSubsF:: (Formula,Formula) -> Map.HashMap Term Term 
          -> Set.HashSet Variable -> Maybe (Map.HashMap Term Term)
buildSubsF (Pred p1 ts1,Pred p2 ts2) m vars = --trace ("buildSubsF1: " ++ show(Pred p1 ts1) ++ show(Pred p2 ts2)) $
  if p1 == p2 && length ts1 == length ts2
  then buildSubsTList (zip ts1 ts2) m vars
  else Nothing
--buildSubsF (Not f1,Not f2) m vars = buildSubsF (f1,f2) m vars
buildSubsF (Not f1@(Pred p1 ts1),Not f2@(Pred p2 ts2)) m vars = --trace ("buildSubsF2: " ++ show(Pred p1 ts1) ++ show(Pred p2 ts2)) $
  if p1 == p2 && length ts1 == length ts2
  then buildSubsTList (zip ts1 ts2) m vars
  else Nothing
buildSubsF _ _ _ = Nothing

buildSubsTList:: [(Term,Term)] -> Map.HashMap Term Term 
              -> Set.HashSet Variable -> Maybe (Map.HashMap Term Term)
buildSubsTList [] acc _ = Just acc
buildSubsTList ((x@(Var vx),y):ts) acc vars
  |x /= y = --trace ("buildSubsTList1: " ++ show x ++ " | " ++ show y ++ show ts) $
    if inVarsOf vx y || Set.member vx vars
    then Nothing
    else case Map.lookup x acc of
      Just term -> if term == y
                   then buildSubsTList ts acc vars
                   else Nothing
      Nothing -> buildSubsTList ts (Map.insert x y acc) vars
  |otherwise = --trace ("buildSubsTList2: " ++ show x ++ " | " ++ show y ++ show ts) $
    buildSubsTList ts acc  vars
buildSubsTList ((Const x,Const y):ts) acc vars
  |x==y = --trace ("buildSubsTList3: " ++ show x ++ " | " ++ show y++ show ts) $ 
    buildSubsTList ts acc vars
  |otherwise = --trace ("buildSubsTList4: " ++ show x ++ " | " ++ show y++ show ts) $ 
    Nothing
buildSubsTList ((Func f1 ts1,Func f2 ts2):ts) acc vars
  |f1 == f2 && length ts1 == length ts2 = --trace ("buildSubsTList5: " ++show (Func f1 ts1) ++ " | " ++ show (Func f2 ts2)++ show ts) $ 
    buildSubsTList (zip ts1 ts2 ++ ts) acc vars
  |otherwise = --trace ("buildSubsTList6: " ++ show (Func f1 ts1) ++ " | " ++ show (Func f2 ts2)++ show ts) $ 
    Nothing
buildSubsTList a _ _= --trace ("Fail:" ++ show a) $
  Nothing

varsOf :: [Formula] -> Set.HashSet Variable
varsOf fs = foldr varsOfF Set.empty fs

varsOfF :: Formula -> Set.HashSet Variable -> Set.HashSet Variable
varsOfF (Pred _ ts) vars = varsOfTList ts vars
varsOfF (Not f) vars = varsOfF f vars
varsOfF (Conn f1 _ f2) vars = 
  Set.unions [vars, (varsOfF f1 Set.empty), (varsOfF f2 Set.empty)]
varsOfF (Quant _ v f) vars = Set.insert v (varsOfF f vars)
varsOfF Box vars = vars

varsOfTList :: [Term] -> Set.HashSet Variable -> Set.HashSet Variable
varsOfTList ts vars = foldr varsOfT vars ts

varsOfT :: Term -> Set.HashSet Variable -> Set.HashSet Variable
varsOfT (Var v) vars = Set.insert v vars
varsOfT (Const _) vars = vars
varsOfT (Func _ ts) vars = varsOfTList ts vars


{- box check -}

boxCheck :: CNFMap -> Bool
boxCheck = Map.member (clauseHash boxClause)

boxClause :: Clause
boxClause = let
  fs = [Box]
  in Clause fs (hash fs) Nothing Nothing

{- Resolution -}

resolutionSteps :: CNFMap -> CNFMap
resolutionSteps cnfmap = 
  trace ("Res:" ++ show (Map.size cnfmap)) -- ++ "\n" ++ cnfMapToString cnfmap ++"\nend") $
  outerloop --combineResults (outerloop cnfmap) cnfmap
  where
    outerloop = foldr (\(c1) !m' -> -- trace (show (Map.size m')) $
      innerloop (fst c1) m') cnfmap sortedCls
    innerloop c1 m = foldr (\c2 !m' -> match c1 (fst c2) m' ) m sortedCls
    match c1 c2 m= let
      fs1 = clauseFormulas c1
      fs2 = clauseFormulas c2
      ls = matchFormulas (fs1,clauseHash c1) (fs2,clauseHash c2) m
      in foldr (\subs !m' -> resolutionStep subs (fs1,c1) (fs2,c2) m') m ls
      
    sortedCls = Map.elems cnfmap 
      -- List.sortBy clauseSize (Map.elems cnfmap)
    clauseSize (Clause fs1 _ _ _, _) (Clause fs2 _ _ _,_)= compare (length fs1) (length fs2)
           

resolutionStep :: ([(Term,Term)],Formula,Formula) -> ([Formula],Clause) -> ([Formula],Clause) ->
                   CNFMap-> CNFMap
resolutionStep (subs,p1,p2) (f1,cl1) (f2,cl2) m = 
  -- trace ("res:"++show cl1 ++ show p1 ++ " with " ++ show cl2 ++ show p2 ++ "subs: " ++ show subs) $
  if not (boxCheck m)
  then
    if null f1' && null f2'
    then -- trace (show (Map.size m) ++" Box found. Slowly finishing process.") $
      Map.insert (clauseHash boxClause) (boxClause{parent1 = Just cl1,parent2 = Just cl2},Set.empty) m
    else
      checkAndInsert fs cl1 cl2 m
  else m
  where
    f1' = map (substituteList subs) (filter (/= p1) f1)
    f2' = map (substituteList subs) (filter (/= p2) f2)
    fs = f1' `List.union` f2'

combineResults :: CNFMap ->CNFMap ->CNFMap
combineResults res old= Map.foldr (\e !m'->comb e m') old res
  where
    comb e@(cl,_) m =
      case (parent1 cl,parent2 cl) of
      (Just p1,Just p2)-> let
        k1 = clauseHash p1
        k2 = clauseHash p2
        m' = ins k1 k2 m
        m'' = ins k2 k1 m'
        in Map.insert (clauseHash cl) e m'' 
      _ -> m
    ins k1 k2 m = Map.adjust (addClauseKey k1) k2 m
    addClauseKey k (cl,keys) = (cl,Set.insert k keys) 


checkAndInsert:: [Formula] -> Clause -> Clause -> CNFMap -> CNFMap
checkAndInsert !fs cl1 cl2 !m = let 
  check0 = not (Map.member fsHash m)
  check1 = not (boxCheck m)
  check2 = not (isTautology fs')
  check3 = not (mapSubsumes fs' m)
  check3' = not (mapCondensateOf fs' m)

  m' = Map.insertWith (sndUnion) (clauseHash cl2) (cl2,singletonKey cl1) m
  m'' = Map.insertWith (sndUnion) (clauseHash cl1) (cl1,singletonKey cl2) m'
--  in if True || (check1 && check2)
  in if check0 && check2 && check3
     then --trace (show cl) $ --(clauseFormulas cl)) $ 
      Map.insert fsHash (cl,Set.empty) m'' --m''
     else m'' -- backwardSubsumption fs m''
  where
    sndUnion (clOld,fsOld) (_,fsNew) = (clOld,Set.union fsOld fsNew)
    singletonKey clause = Set.singleton (clauseHash clause)
    fs' = if length fs == 1
          then normalizeFormulas ('u':show (Map.size m) ++['_']) 1 fs
          else fs
    fsHash = hash fs'
    cl = Clause fs' fsHash (Just cl1) (Just cl2)
 
matchFormulas :: ([Formula],Int) -> ([Formula],Int) -> CNFMap -> [([(Term,Term)],(Formula),(Formula))]
matchFormulas (f1,h1) (f2,h2) cnfmap =
  if lookupFormula h1 h2 cnfmap
  then []
  else loop1 f1 f2 --`List.union` loop1 f2 f1
  where
    loop1 ps qs = foldr (\p res -> (loop2 p qs) ++ res) [] ps
    loop2 p qs = foldr (\q res -> func2 p q res) [] qs
    func2 p q res =
      case unificatorOpposite p q of
        Just tt -> (tt,(p),(q)): res
        Nothing -> res
    lookupFormula h1' h2' m =
      case Map.lookup h1' m of
        Just (_,s) -> Set.member h2' s
        Nothing -> False


normalizeFormulas :: String -> Int -> [Formula] -> [Formula]
normalizeFormulas chs i fs = 
  let (_,_,fs') = foldr (normForm chs) (i,Map.empty,[]) fs 
  in fs'
  where
    normForm chars f (i,!m,fs') = let
      (i',m',f') = normalizeFormula chars (i,m,f)
      in (i',m',f':fs')
    
normalizeFormula :: String -> (Int,Map.HashMap Term Term,Formula) -> (Int,Map.HashMap Term Term,Formula)
normalizeFormula chs (i,m,Pred p ts) = 
  let
    (i',m',ts') = normalizeTerms chs (i,m,ts)
  in (i',m',Pred p ts')
normalizeFormula chs (i,m,Not f) = 
  let
    (i',m',f') = normalizeFormula chs (i,m,f)
  in (i',m',Not f')
normalizeFormula chs (i,m,Quant q v f) =
  let
    var = Var v
    origVar = Map.lookup var m
    nv = chs ++ show i
    var' = Var nv
    m' = Map.insert var var' m
    (i',m'',f') = normalizeFormula chs (i+1,m',f)
  in case origVar of
     Nothing -> (i',m'',Quant q nv f')
     Just v'@(Var ov) -> (i',Map.insert var v' m'',Quant q ov f')
normalizeFormula chs (i,m,Conn f1 c f2) =
  let
    (i',m',f1') = normalizeFormula chs (i,m,f1)
    (i'',m'',f2') = normalizeFormula chs (i',m',f2)
  in (i'',m'',Conn f1' c f2')

normalizeTerms :: String -> (Int,Map.HashMap Term Term,[Term]) -> (Int,Map.HashMap Term Term,[Term])
normalizeTerms chs (i,m,ts) = foldr (normTerm chs) (i,m,[]) ts 
  where
    normTerm chars t (i,m,ts') = let
      (i',m',t') = normalizeTerm chars (i,m,t)
      in (i',m',t':ts')

normalizeTerm :: String -> (Int,Map.HashMap Term Term,Term) -> (Int,Map.HashMap Term Term,Term)
normalizeTerm chs (i,m,var@(Var _)) =
  case Map.lookup var m of
    Just var' -> (i,m,var')
    Nothing -> let
      var' = Var (chs ++ show i)
      i1 = i+1
      m' = Map.insert var var' m
      in (i1,m',var')
normalizeTerm chs orig@(_,_,Const _) = orig
normalizeTerm chs (i,m,(Func f ts)) =
  let
    (i',m',ts') = normalizeTerms chs (i,m,ts)
  in (i',m',Func f ts')

  
{- tautology removal -}

--stripTautologies :: CNFMap -> CNFMap
--stripTautologies m = Map.filterWithKey check m
--  where
--    check _ (cl,_) = not (isTautology cl)

isTautology :: [Formula] -> Bool
isTautology preds = 
  loop part
  where
    loop (p:ps) = or (map p preds) || loop ps
    loop [] = False
    part = map (oppositeForm) preds
    oppositeForm p1@(Pred _ _) (Not p2@(Pred _ _)) =
      p1 == p2
    oppositeForm (Not p1@(Pred _ _)) p2@(Pred _ _) =
      p1 == p2
    oppositeForm _ _ = False
    -- oppositeForm p1@(Pred _ _) (Not p2@(Pred _ _)) =
    --   case unificator p1 p2 of
    --     Just _  -> True
    --     Nothing -> False
    -- oppositeForm (Not p1@(Pred _ _)) p2@(Pred _ _) =
    --   case unificator p1 p2 of
    --     Just _  -> True
    --     Nothing -> False
    -- oppositeForm _ _ = False


{- unification -}

unificatorOpposite :: Formula -> Formula -> Maybe [(Term,Term)]
unificatorOpposite p1@(Pred _ _) (Not p2@(Pred _ _)) = unificator p1 p2
unificatorOpposite (Not p1@(Pred _ _)) p2@(Pred _ _) = unificator p1 p2
unificatorOpposite _ _ = Nothing

unificator :: Formula -> Formula -> Maybe [(Term,Term)]
unificator = unificatorF

unificatorF :: Formula -> Formula -> Maybe [(Term,Term)]
unificatorF (Not f1@(Pred _ _)) (Not f2@(Pred _ _)) =
  unificatorF f1 f2
unificatorF (Pred p1 ts1) (Pred p2 ts2) =
  if p1 == p2 && length ts1 == length ts2
  then
    unificatorT (zip ts1 ts2)
  else
    Nothing
unificatorF _ _ = Nothing

unificatorT :: [(Term,Term)] -> Maybe [(Term,Term)]
unificatorT acc =
  case acc' of
    Just x ->
      if x == acc
      then Just x
      else unificatorT x
    Nothing -> Nothing
  where
    acc' = unificatorT' acc []

unificatorT' :: [(Term,Term)] -> [(Term,Term)] -> Maybe [(Term,Term)]
unificatorT' ((Const c1,Const c2):xs) acc =
  if c1 == c2
  then unificatorT' xs (acc)
  else Nothing
unificatorT' ((Func f1 ts1,Func f2 ts2):xs) acc =
  if f1 == f2 && length ts1 == length ts2
  then unificatorT' (xs ++ zip ts1 ts2) acc
  else Nothing
unificatorT' ((Const c,Var v):xs) acc =
  unificatorT' ((Var v,Const c):xs) acc
unificatorT' ((Func f ts,Var v):xs) acc =
  unificatorT' ((Var v,Func f ts):xs) acc
unificatorT' (t@(Var v1,Var v2):xs) acc =
  if v1 == v2
  then unificatorT' xs (acc)
  else if or (map (inVarsOfTT v1) xs) || or (map (inVarsOfTT v1) acc)
       then unificatorT' xs' (t:acc')
       else unificatorT' xs (t:acc)
  where
    xs' = map (substituteTT t) xs
    acc' = map (substituteTT t) acc
unificatorT' (t@(Var v,Const _):xs) acc =
  if or (map (inVarsOfTT v) xs) || or (map (inVarsOfTT v) acc)
  then unificatorT' xs' (t:acc')
  else unificatorT' xs (t:acc)
  where
    xs' = map (substituteTT t) xs
    acc' = map (substituteTT t) acc
unificatorT' (t@(Var v,fn@(Func _ _)):xs) acc =
  if ((or (map (inVarsOfTT v) xs))
      || (or (map (inVarsOfTT v) acc)))
     && not inFunc
  then unificatorT' xs' (t:acc')
  else if inFunc
       then Nothing
       else unificatorT' xs (t:acc)
  where
    inFunc = inVarsOf v fn
    xs' = map (substituteTT t) xs
    acc' = map (substituteTT t) acc

unificatorT' [] acc = Just (reverse acc)
unificatorT' _ _ = Nothing
--unificatorT' xs ys = error ("xs:"++ show xs ++ " ys:" ++ show ys)

inVarsOfTT :: Variable -> (Term,Term) -> Bool
inVarsOfTT v (x,y) = inVarsOf v x || inVarsOf v y

inVarsOfF :: Variable -> Formula -> Bool
inVarsOfF v (Pred _ ts) = any (inVarsOf v) ts
inVarsOfF v (Not f) = inVarsOfF v f
inVarsOfF v (Conn f1 _ f2) = inVarsOfF v f1 || inVarsOfF v f2
inVarsOfF v (Quant _ var f) = var == v || inVarsOfF v f
inVarsOfF _ Box = False


inVarsOf :: Variable -> Term -> Bool
inVarsOf v (Var x) = v == x
inVarsOf _ (Const _) = False
inVarsOf v (Func _ ts) = any (inVarsOf v) ts



substituteList :: [(Term,Term)] -> Formula -> Formula
substituteList tts f = foldr (\s f' -> substituteF s f') f subs
  where
    subs = filter isSub tts
    isSub (Var _, _) = True
    isSub _ = False

substituteF :: (Term,Term) -> Formula -> Formula
substituteF subs (Pred p ts) = Pred p (map (substituteT subs) ts)
substituteF subs (Not f) = Not (substituteF subs f)
substituteF subs (Conn f1 c f2) = Conn (substituteF subs f1) c (substituteF subs f2)
substituteF subs@(Var v,_) (Quant q var f)
  | var == v = Quant q var f
  | otherwise = Quant q var (substituteF subs f)
substituteF subs (Quant _ _ _) = error ("substitution "++show subs++" not valid\n")
substituteF _ Box = Box  
  
substituteTT :: (Term,Term) -> (Term,Term) -> (Term,Term)
substituteTT subs (t1,t2) = (substituteT subs t1,substituteT subs t2)

substituteT :: (Term,Term) -> Term -> Term
substituteT _ c@(Const _) = c
substituteT (Var v,t) (Var v') =
  if v == v'
  then t
  else Var v'
substituteT subs (Func f ts) = Func f (map (substituteT subs) ts)
substituteT subs _ = error ("substitution "++show subs++" not valid\n")



{- build cnf map -}


cnf :: Formula -> CNFMap
cnf f = let
  disjClauses = map disjunctsToClause . splitAnds . stripForAlls . cnfF
  cnfMap = foldr func Map.empty (disjClauses f)
  in cnfMap
  where
    func c m = Map.insert (hash c) (c,Set.empty) m

cnfF :: Formula -> Formula
cnfF = distOr . moveQuant {- . addConstantEqualities -} . rssk . nnf . rectify

stripForAlls :: Formula -> Formula
stripForAlls (Quant ForAll _ f) = stripForAlls f
stripForAlls f@(Conn _ And _) = f
stripForAlls f@(Conn _ Or _) = f
stripForAlls f@(Pred _ _) = f
stripForAlls f@(Not (Pred _ _)) = f
stripForAlls f = error ("stripForAlls: Formula not in correct form. " ++
                        show f)

splitAnds :: Formula -> [Formula]
splitAnds (Conn f1 And f2) = splitAnds f1 ++ splitAnds f2
splitAnds f@(Conn _ Or _) = [f]
splitAnds f@(Pred _ _) = [f]
splitAnds f@(Not (Pred _ _)) = [f]
splitAnds f = error ("splitAnds: Formula not in correct form. " ++ show f)

disjunctsToClause :: Formula -> Clause
disjunctsToClause preds = let
  fs = Set.toList . disjunctsToSet $ preds
  in Clause fs (hash fs) Nothing Nothing

disjunctsToSet :: Formula -> Set.HashSet Formula
disjunctsToSet (Conn f1 Or f2) =
  Set.union (disjunctsToSet f1) (disjunctsToSet f2)
disjunctsToSet f@(Pred _ _) = Set.singleton f
disjunctsToSet f@(Not (Pred _ _)) = Set.singleton f
disjunctsToSet f = error ("disjunctsToSet: Formula not in correct form. " ++ show f)



{- distribute over Or -}

distOr :: Formula -> Formula
distOr = distOrF

distOrF :: Formula -> Formula
distOrF p@(Pred _ _) = p
distOrF f@(Not (Pred _ _)) = f
distOrF (Not (Not f)) = distOrF f
distOrF (Conn (Conn f1 And f2) Or c2) = let
  c2' = distOrF c2
  l = distOrF (Conn f1 Or c2')
  r = distOrF (Conn f2 Or c2')
  in (Conn l And r)
distOrF (Conn c1 Or (Conn f3 And f4)) = let
  c1' = distOrF c1
  l = distOrF (Conn c1' Or f3)
  r = distOrF (Conn c1' Or f4)
  in (Conn l And r)
distOrF f@(Conn f1 c f2) = let
  f' = Conn (distOr f1) c (distOr f2)
  in if f == f'
     then
       f
     else
       distOrF f'
distOrF (Quant q v f) = Quant q v (distOrF f)
distOrF Box = Box
distOrF e = error $ show e

{- move quantifiers to top -}

moveQuant :: Formula -> Formula
moveQuant p@(Pred _ _) = p
moveQuant (Not (Quant ForAll v f)) = Quant Exists v (moveQuant (Not f))
moveQuant (Not (Quant Exists v f)) = Quant ForAll v (moveQuant (Not f))
moveQuant (Conn (Quant q v f) c f2) = Quant q v (moveQuant (Conn f c f2))
moveQuant (Conn f1 c (Quant q v f)) = Quant q v (moveQuant (Conn f1 c f))
moveQuant f@(Conn f1 c f2) = let
  f' = Conn (moveQuant f1) c (moveQuant f2)
  in if f == f'
     then
       f
     else
       moveQuant f'
moveQuant (Not f) = Not f
moveQuant (Quant q v f) = Quant q v (moveQuant f)
moveQuant Box = Box
-- moveQuant e = error $ show e

{- refutational strutural skolemization -}

rssk :: Formula -> Formula
rssk = fst . (rsskF "f" 0 [] Map.empty)

rsskF :: String -> Int -> [Variable] -> Map.HashMap Variable Term -> Formula -> (Formula,Int)
rsskF _ i _ m (Pred p ts) = (Pred p (map (rsskT m) ts),i)
rsskF base i deps m (Not f) = let
  (f',i') = rsskF base i deps m f
  in (Not f',i')
rsskF base i deps m (Conn f1 c f2) = let
  (f1',i')  = rsskF base i deps m f1
  (f2',i'') = rsskF base i' deps m f2
  in (Conn f1' c f2',i'')
rsskF base i deps m (Quant ForAll v f) = let
  deps' = v:deps
  (f',i') = rsskF base i deps' m f
  in (Quant ForAll v f',i')
rsskF base i deps m (Quant Exists v f) = let
  func = if null deps
         then Const (v ++ base ++ show i)
         else Func (v ++ base ++ show i) (map (Var) (reverse deps))
  m' = Map.insert v func m
  in rsskF base (i+1) deps m' f
rsskF _ i _ _ Box = (Box,i)

rsskT :: Map.HashMap Variable Term -> Term -> Term
rsskT m v@(Var x) =
  case Map.lookup x m of
    Just func -> func
    Nothing   -> v
rsskT _ c@(Const _) = c
rsskT m (Func f ts) = Func f (map (rsskT m) ts)

{- negation normal form -}

nnf :: Formula -> Formula
nnf = nnfF

nnfF :: Formula -> Formula
nnfF (Not (Quant ForAll v f)) = Quant Exists v (nnfF (Not f))
nnfF (Not (Quant Exists v f)) = Quant ForAll v (nnfF (Not f))
nnfF (Not (Not f)) = nnfF f
nnfF (Not (Conn f1 And f2)) = Conn (nnfF (Not f1)) Or (nnfF (Not f2))
nnfF (Not (Conn f1 Or f2)) = Conn (nnfF (Not f1)) And (nnfF (Not f2))
nnfF (Not (Conn f1 Imp f2)) = Conn (nnfF f1) And (nnfF (Not f2))
nnfF p@(Pred _ _) = p
nnfF n@(Not (Pred _ _)) = n
nnfF (Conn f1 Imp f2) = Conn (nnfF (Not f1)) Or (nnfF f2)
nnfF (Conn f1 c f2) = Conn (nnfF f1) c (nnfF f2)
nnfF (Quant q v f) = Quant q v (nnfF f)
nnfF Box = Box
nnfF (Not Box) = Not Box

{- rectification -}

rectify :: Formula -> Formula
rectify = fst . (rectifyF "n" 0 Map.empty)

rectifyF :: String -> Int -> Map.HashMap Variable Variable -> Formula -> (Formula,Int)
rectifyF _ i m (Pred p ts) = (Pred p (map (rectifyT m) ts),i)
rectifyF base i m (Not f) = let
  (f',i') = rectifyF base i m f
  in (Not f', i')
rectifyF base i m (Conn f1 c f2) = let
  (f1',i')  = rectifyF base i m f1
  (f2',i'') = rectifyF base i' m f2
  in (Conn f1' c f2' ,i'')
rectifyF base i m (Quant q v f) = let
  val = base ++ show i
  m' = Map.insert v val m
  (f',i') = rectifyF base (i+1) m' f
  in (Quant q val f',i')
rectifyF _ i _ Box = (Box,i)

rectifyT :: Map.HashMap String String -> Term -> Term
rectifyT _ c@(Const _) = c
rectifyT m v@(Var x) =
  case Map.lookup x m of
    Just val -> Var val
    Nothing  -> v
rectifyT m (Func f ts) = Func f (map (rectifyT m) ts)

{- find constants -}

addConstantEqualities :: Formula -> Formula
addConstantEqualities f = 
  if Set.null consts
  then f
  else Conn eqs And f
  where
    consts = getConstants f
    eqs = buildEqualityPredicates consts

buildEqualityPredicates :: Set.HashSet Constant -> Formula
buildEqualityPredicates cs
  | otherwise = (foldr (\c f -> Conn f And (eqPred c)) first rest)
  where
    eqPred c = Pred equalPredicate [c,c]
    list = map Const (Set.toList cs)
    first = eqPred (head list)
    rest = tail list

addEqualityAxioms :: CNFMap -> CNFMap
addEqualityAxioms !m = trace "one" $ let
  m' = insertKeys ks cnfSymm m
  in insertKeys ks cnfTrans m'
  where
    cnfSymm = fst . head $ (Map.elems (cnf equalityAxiomSymmetry))
    cnfTrans = fst . head $ (Map.elems (cnf equalityAxiomTransitivity))
    --reflex = Quant ForAll "x" (Pred equalPredicate [(Var "x"),(Var "x")])
    ks = Set.fromList [clauseHash cnfSymm,clauseHash cnfTrans]
    insertKeys k cl mp = Map.insert (clauseHash cl) (cl,k) mp


equalityAxiomSymmetry = Quant ForAll "x" (Quant ForAll "y" (
      Conn (Pred equalPredicate [(Var "x"),(Var "y")]) Imp
           (Pred equalPredicate [(Var "y"),(Var "x")])
      ))
equalityAxiomTransitivity = Quant ForAll "x" (Quant ForAll "y" ( Quant ForAll "z" (
      Conn (Conn (Pred equalPredicate [(Var "x"),(Var "y")]) And
                 (Pred equalPredicate [(Var "y"),(Var "z")])) Imp
           (Pred equalPredicate [Var "x",Var "z"])
      )))
      
getConstants :: Formula -> Set.HashSet Constant
getConstants = getConstantsF

getConstantsF :: Formula -> Set.HashSet Constant
getConstantsF (Pred _ ts) = foldr Set.union Set.empty (map getConstantsT ts)
getConstantsF (Quant _ _ f) = getConstantsF f
getConstantsF (Conn f1 _ f2) = (getConstantsF f1) `Set.union` (getConstantsF f2)
getConstantsF (Not f) = getConstantsF f
getConstantsF Box = Set.empty

getConstantsT :: Term -> Set.HashSet Constant
getConstantsT (Var _) = Set.empty
getConstantsT (Const c) = Set.singleton c
getConstantsT (Func _ ts) = foldr Set.union Set.empty (map getConstantsT ts)

